from .cBMG_tools import analyze_digraphs_list
from .common_tools import norm_path
import pandas as pd

####################
#                  #
# Standalone usage #
#                  #
####################

if __name__ == "__main__":

    import argparse
    import os
    from .parse_prt import parse_prt_project_bmgs
    from .in_out import read_tl_digraph

    parser = argparse.ArgumentParser(prog= 'is_cBMG',
                                     description='Determines if a directed graph is a coloured Best Match Graph (cBMG).',
                                    )
    # Parameters for input graph
    ############################

    parser.add_argument('edges_list',
                        type= str,
                        #required=False,
                        help= '.tsv file containing directed edges.'
                       )

    parser.add_argument('-F', '--edges_format',
                        type= str,
                        required=False,
                        help= 'Format of the edges list (default "prt"). For more information see:...',
                        choices= ['prt', 'tl'], # 'prt_pre_parse'
                        default= 'tl',
                       )

    parser.add_argument('-o', '--output_prefix',
                        type= str,
                        required=False,
                        help= 'prefix used for output files (default "tl_iscBMG").',
                        default= 'tl_iscBMG',
                       )

    args= parser.parse_args()

    # Input data
    #############
    print('Reading hit list...')
    if args.edges_format=='prt':
        args.edges_list= norm_path(args.edges_list)
        #> Añadir opción para guardar o no tabla parseada
        G= parse_prt_project_bmgs(args.edges_list, f_value= 0.9)
    elif args.edges_format=='tl':
        G= read_tl_digraph(args.edges_list, og_col= 'OG')
    else:
        #> Leer tabla de aristas parseada
        raise ValueError(f'Edges format "{args.edges_format}" not valid.')

    # Analyze
    #########
    print('Analyzing hit graphs...')
    df= analyze_digraphs_list(G)
    opath= args.output_prefix + '.csv'
    df.to_csv(opath, sep='\t', index= False)
    print(f'Output successfully writen to {opath}')




from . import nxTree as nxt
from .triplets import get_triplets
from .build_st_n import BUILDst_N
from .nhxx_tools import read_nhxx, get_nhx

import networkx as nx
import pandas as pd

def reconciliate(Tg, Ts, gene_attr= 'genes', species_attr= 'species', event_attr= 'event'):
    """
    The input trees must be planted, where the LCA of all the leafs have int label '1'
    """
    # Check if species tree displays the gene tree
    Rg= get_triplets(Tg, event=event_attr, color= species_attr) + [ nxt.subtree(Ts, 1) ]
    Trr= BUILDst_N(Rg, color_attr= species_attr)
    # Obtain reconciliation map
    mu= get_reconciliation_map(Tg, Ts, event_attr= event_attr, species_attr= species_attr)
    # Resolve gene tree
    T, mu_s, mu_inv= resolve(Tg, Ts, mu, event_attr, [species_attr], Tg_root= 1)
    return T, mu_s, mu_inv

def resolve(Tg, Ts, mu, event_attr, copy_attrs= [], Tg_root= None):
    # Initialize output objects
    if Tg_root==None:
        Tg_root= Tg.root
    T= nx.DiGraph()
    T.root= 0
    T.add_node(0, **{event_attr: 0})
    n= 1
    dad_dict= {Tg_root : 0}
    mu_s= {}
    mu_inv= {y:{'S':[],'P':[],'X':[],'.':[]} for y in Ts}
    # Given is a universal parameter for all functions
    given= Tg, Ts, T, mu, event_attr, copy_attrs, mu_s, mu_inv, dad_dict
    # Segregate all the nodes
    for x0 in nx.dfs_preorder_nodes(Tg, Tg_root):
        symbol= Tg.nodes[x0][event_attr]
        if symbol=='S':
            resolve_node= _resolve_s
        elif symbol=='P':
            resolve_node= _resolve_p
        elif nxt.is_leaf(Tg, x0):
            resolve_node= _resolve_l
        else:
            raise ValueError(f'unknown event "{symbol}"')
        n= resolve_node(x0, n, dad_dict[x0], *given)
    return T, mu_s, mu_inv

def get_reconciliation_map(Tg, Ts, event_attr= 'event', species_attr= 'species'):
    mu= {Tg.root: Ts.root}
    Q= nxt.BFS_graph(Tg, Tg.root)
    Q.pop(0)
    for x in Q:

        if nxt.is_leaf(Tg, x):
            y= nxt.LCA_color(Ts, {Tg.nodes[x][species_attr]}, color_attr= 'species')

        else:
            x_species= nxt.induced_colors(Tg, x, color_attr= species_attr)
            y= nxt.LCA_color(Ts, x_species, color_attr= 'species')

        mu[x]= y
    return mu

def recon_2_text(mu):
    text_mu= 'Tg_node\tTs_node\n'
    text_mu+= '\n'.join(('\t'.join(map(str, x)) for x in mu.items()))
    return text_mu

##################
#                #
# Internal usage #
#                #
##################

def _resolve_s(x0, n, dad, *given):
    # Obtain constant arguments
    Tg, Ts, T, mu, event_attr, copy_attrs, mu_s, mu_inv, dad_dict= given
    # Resolve
    L= {mu[x1] for x1 in Tg[x0]}
    d, Xi, Kappa= _inherit(mu[x0], L, n, dad, *given)##################### Quitar Xi
    # Update dad_dict
    for x1 in Tg[x0]:
        _update_ddict(x0, x1, *given)
    return n+d

def _resolve_p(x0, n, dad, *given):
    Tg, Ts, T, mu, event_attr, copy_attrs, mu_s, mu_inv, dad_dict= given
    # Resolve
    y0= mu[x0]
    _add_node(n, 'P', y0, dad, *given)
    d= 0
    n1= n+1
    for x1 in Tg[x0]:
        y1= mu[x1]
        if y0==y1:
            dad_dict[x1]= n
        else:
            d, Xi, Kappa= _inherit(y0, {y1}, n1, n, *given)
            n1+= d
            _update_ddict(x0, x1, *given)

    return n1

def _resolve_l(x0, n, dad, *given):
    Tg, Ts, T, mu, event_attr, copy_attrs, mu_s, mu_inv, dad_dict= given
    _add_leaf(n, Tg.nodes[x0][event_attr], mu[x0], dad, *given)
    T.nodes[n].update({attr : Tg.nodes[x0][attr] for attr in copy_attrs})
    return n+1

def _update_ddict(x0, x1, *given):
    Tg, Ts, T, mu, event_attr, copy_attrs, mu_s, mu_inv, dad_dict= given
    y= mu[x1]
    if y!=mu[x0]:
        y= nxt.get_dad(Ts, mu[x1])
    dad_dict[x1]= mu_inv[y]['S'][-1]

def _inherit(roh, lamb, n, dad, *given):
    """
    _inherit(roh, lamb, n, dad, *given)

    Simulates the process of inheriting a gene throug a sub-species tree defined by roh and lamb;
    rho is a node in the species tree, and lamb is a list of nodes y in the species tree such that
    rho >= y.
    The subtree is the subgraph of Ts induced by all the nodes y' such that 
    roh >= y' > y,
    and also the node roh.

    This function creates the node n on the genes tree as son of dad.
    Creates all the necessary descendants of n, usign as a guide the sub-species tree defined by roh and lab.
    Such descendants may be of type speciation and loss.

    Input:
    - roh: Node of the species tree where the process starts.
    - lamb: Set of nodes {y0, y1, ...} of the species tree where the process ends.
    - n: Nodo to create in the gene tree, corresponds to the speciation at roh.
    - dad: dad of n in the genes tree.
    """
    Tg, Ts, T, mu, event_attr, copy_attrs, mu_s, mu_inv, dad_dict= given
    # Inherit
    d= 0
    Xi, Kappa= _ancestry_graph(roh, lamb, *given)
    Q= [(roh, dad)]
    while len(Q)>0:
        n1= n+d
        y, dad= Q.pop(0)
        _add_node(n1, 'S', y, dad, *given)
        d+= 1 + _add_losses(y, Xi, n1, *given)
        Q+= [(y1, n1) for y1 in Kappa[y]]
    return d, Xi, Kappa

def _add_node(n, phi, y, dad, *given):
    """
    Agrega un nodo al árbol de genes, y guarda su información.

    Input:
    - n: nodo a crear.
    - phi: símbolo del nodo ('S', 'P', 'X', '.').
    - y: Nodo al que mapea n por el mapeo mu_s.
    - dad: Papa de n.
    """
    Tg, Ts, T, mu, event_attr, copy_attrs, mu_s, mu_inv, dad_dict= given
    # Add node and edge
    T.add_node(n, **{event_attr: phi})
    T.add_edge(dad, n)
    mu_s[n]= y
    mu_inv[y][phi]+= [n]

def _add_leaf(n, label, y, dad, *given):
    Tg, Ts, T, mu, event_attr, copy_attrs, mu_s, mu_inv, dad_dict= given
    # Add node and edge
    T.add_node(n, **{event_attr: label})
    T.add_edge(dad, n)
    mu_s[n]= y
    mu_inv[y]['.']+= [n]

def _ancestry_graph(y0, lamb, *given):
    Tg, Ts, T, mu, event_attr, copy_attrs, mu_s, mu_inv, dad_dict= given
    Kappa_nodes= {y0}
    Q= list(lamb-Kappa_nodes)
    while len(Q)>0:
        y= Q.pop(0)
        y_dad= nxt.get_dad(Ts, y)
        if y_dad not in Kappa_nodes:
            Kappa_nodes.add(y_dad)
            Q+= [y_dad]
    Xi_nodes= Kappa_nodes.union(lamb)
    # Create graphs
    Kappa= nx.induced_subgraph(Ts, Kappa_nodes)
    Xi= nx.induced_subgraph(Ts, Xi_nodes)
    return Xi, Kappa

def _add_losses(y, Xi, n, *given):
    Tg, Ts, T, mu, event_attr, copy_attrs, mu_s, mu_inv, dad_dict= given
    d= 0
    for y1 in set(Ts[y]) - set(Xi[y]):
        _add_node(n+1+d, 'X', y1, n, *given)
        d+= 1
    return d

####################
#                  #
# Standalone usage #
#                  #
####################

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(prog= 'reconciliate',
                                     description='Maps and resolves gene trees against species trees',
                                    )
    # Parameters for input graph
    ############################

    parser.add_argument('gene_trees',
                        type= str,
                        help= 'file containing gene trees in newick format.'
                       )

    parser.add_argument('species_tree',
                        type= str,
                        help= 'file containing a species tree in newick format.'
                       )

    parser.add_argument('-o', '--output_prefix',
                        type= str,
                        required=False,
                        help= 'prefix used for output files (default "tl_project").',
                        default= 'tl_project',
                       )

    args= parser.parse_args()

    # Process data
    ##############
    with open (args.gene_trees) as F:
        Tg= read_nhxx(F.readline().strip())
    with open (args.species_tree) as F:
        Ts= read_nhxx(F.readline().strip(), name_attr= 'species')

    Tr, mu, mu_inv= reconciliate(Tg, Ts, gene_attr= 'genes',
                                 species_attr= 'species',
                                 event_attr= 'label',
                                )

    nhx_r, text_mu= recon_2_text(Tr, mu)

    with open(f'{args.output_prefix}.resolved_tree.nhx', 'w') as F:
        F.write(nhx_r)
    with open(f'{args.output_prefix}.reconciliation_map.nhx', 'w') as F:
        F.write(text_mu)

    keys= ['S', 'P', 'X', '.']
    aux= pd.Series(mu_inv).apply(lambda X: pd.Series([','.join(map(str, X[k])) for k in keys])).reset_index()
    aux.columns= ['Ts_node'] + keys
    aux.to_csv(f'{args.output_prefix}.inverse_map.nhx' , sep= '\t', index= False)



need_params= {0: ['prt_path'],
              1: ['hit_list'],
              2: [],
             }

def check_args(args):
    d_args= vars(args)
    s0, s1= sorted(args.steps)
    missed= {}
    for step in [s0]:
        for param in need_params[step]:
            if d_args[param] == None:
                missed[step]= missed.get(step, []) + [param]

    if len(missed)>0:
        raise MissedData(f'{missed}')

    return s0, s1

def assign_id(T, idAttr= 'node_id'):
    for x in T:
        T.nodes[x][idAttr]= x


from .reconciliation import reconciliate, recon_2_text
from .error import InconsistentTrees
from tqdm import tqdm

def aux(Tg, Ts):
    try:
        Tr, mu_s, mu_inv= reconciliate(Tg, Ts, gene_attr= 'genes', species_attr= 'species', event_attr= 'label')
        return Tr, mu_s
    except InconsistentTrees:
        return '*'

def reconciliate_many(gTrees, Ts):
    recons= [aux(Tg, Ts) for Tg in tqdm(gTrees)]
    return recons

if __name__ == "__main__":
    from .in_out import read_tl_digraph, write_digraphs_list
    from .constants import _df_matches_cols, _default_f
    from .hug_free import build_graph, get_augmented_tree
    from .error import MissedData, ParameterError
    from .parse_prt import _parse
    from .common_tools import norm_path
    from .cBMG_tools import analyze_digraphs_list
    from .nhxx_tools import read_nhxx, get_nhx
    import pandas as pd
    import argparse
    import os

    parser = argparse.ArgumentParser(prog= 'revolutionhtl',
                                     description='Bioinformatics tool for the reconstruction of evolutionaty histories.',
                                    )
    # Parameters
    ############

    parser.add_argument('-steps',
                        help= 'list of steps to run (default: 1 2 3).',
                        type= int,
                        nargs= '*',
                        default= [1, 2, 3]
                       )

    parser.add_argument('-prt_path',
                        help= 'path to a directory containing proteinortho output files.',
                        type= str,
                        default = './',
                       )

    parser.add_argument('-gene_trees',
                        help= '.tsv file containing a .nhx for each line at column "tree"',
                        type= str,
                       )

    parser.add_argument('-species_tree',
                        help= '.nhx file containing a species tree.',
                        type= str,
                       )

    parser.add_argument('-hit_list',
                        help= '.tsv file containing hits.',
                        type= str,
                       )

    parser.add_argument('-og', '--orthogroup_column',
                        help= 'column in -hit_list and -gene_trees specifying orthogroups (default: OG).',
                        type= str,
                        default= 'OG',
                       )

    parser.add_argument('-o', '--output_prefix',
                        help= 'prefix used for output files (default "tl_project").',
                        type= str,
                        default= 'tl_project',
                       )

    parser.add_argument('-rod', '--recon_output_dir',
                        help= 'directory for reconciliation maps.',
                        type= str,
                        default= './',
                       )

    parser.add_argument('-f', '--f_value',
                        help= f'number between 0 and 1 used for the adaptative threshhold for best matches selection: f*max_bit_score. (default {_default_f}, see proteinortho paper for a deep explanation)',
                        type= float,
                        default= _default_f,
                       )

    args= parser.parse_args()

    if not args.recon_output_dir.endswith('/'):
        args.recon_output_dir+= '/'

    ################
    # Process data #
    ################

    print('\nREvolutionH-tl')

    allowed_steps= {0, 1, 2, 3}
    bad_steps= set(args.steps) - allowed_steps
    if len(bad_steps) > 0:
        raise ParameterError(f'Only steps 0, 1, 2, and 3 are allowed to be used in the parameter -steps. ')
    else:
        args.steps= sorted(set(args.steps))

    print(f'Running steps {", ".join(map(str, args.steps))}')

    # 0. Convert proteinortho output to best hit list
    #################################################
    if 0 in args.steps:
        print("\nStep 0: Convert proteinortho output to best hit list")
        print("----------------------------------------------------")
        if args.prt_path==None:
            raise MissedData('Step 0 needs a value for the parameter -prt_path')

        args.prt_path= norm_path(args.prt_path)
        opath= args.output_prefix+'.best_hits.tsv'
        _parse(args.prt_path, args.f_value, opath)
        print('This file will be used as input for step 1.')
        args.hit_list= opath

    # 1. Conver best hit graphs to cBMGs
    ####################################
    if 1 in args.steps:
        print("\nStep 1: Conver best hit graphs to cBMGs")
        print("---------------------------------------")
        if args.hit_list==None:
            raise MissedData('Step 1 needs a value for the parameter -hit_list')

        print('Reading hit graphs...')
        G= read_tl_digraph(args.hit_list, og_col= args.orthogroup_column)
        print('Editing to best match graphs (cBMGs)...')
        G1= G.progress_apply(lambda X: build_graph(X, color_attr= 'species'))
        opath= f'{args.output_prefix}.cBMGs.tsv'
        write_digraphs_list(G1, opath)#, n_attrs= ['gene', 'species'])
        print(f'Best match graphs successfully writen to {opath}')
        print('This file will be used as input of step 2.')
        args.hit_list= opath

    # 2. Conver cBMGs to gene trees
    ###############################
    if 2 in args.steps:
        print("\nStep 2: Reconstruct gene trees")
        print("------------------------------")
        if args.hit_list==None:
            raise MissedData('Step 2 needs a value for the parameter -hit_list')

        print('Reading best match graphs...')
        G= read_tl_digraph(args.hit_list, og_col= args.orthogroup_column)

        print('Reconstructing gene trees...')
        df= analyze_digraphs_list(G, species_attr= 'species', gene_attr= 'accession')

        print('Labeling gene tree nodes with evolutionary events...')
        df.tree= df.tree.progress_apply( lambda x: get_augmented_tree(x, 'accession', 'species') )
        df.tree= df.tree.apply(lambda x: get_nhx(x, root= 1, name_attr= 'accession'))

        opath= f'{args.output_prefix}.gene_trees.tsv'
        df.to_csv(opath, sep='\t', index= False)
        print(f'Gene trees successfully writen to {opath}')
        print('This file will be used as input of step 3.')
        args.gene_trees= opath

    # 3. Reconciliate gene trees and species tree
    ##############################################
    if 3 in args.steps:
        print("\nStep 3: Reconciliate gene and species trees")
        print("-------------------------------------------")
        if args.gene_trees==None:
            raise MissedData('Step 2 needs a value for the parameter -gene_trees')
        if args.species_tree==None:
            raise MissedData('Step 2 needs a value for the parameter -species_tree')
        if not os.path.isdir(args.recon_output_dir):
            raise ValueError(f'-recon_output_dir {args.recon_output_dir}: : No such directory')

        print('Reading trees...')
        gTrees= pd.read_csv(args.gene_trees, sep= '\t').set_index(args.orthogroup_column).tree.apply(read_nhxx)
        with open(args.species_tree) as F:
            sTree= read_nhxx(''.join( F.read().strip().split('\n') ),
                             name_attr= 'species')
            assign_id(sTree)

        print('Reconciliating trees...')
        recons= reconciliate_many(gTrees, sTree)

        # Write resolved trees
        df_r= pd.DataFrame([(idx, X) if X == '*' else (idx, X[0])
                            for idx, X in zip(gTrees.index, recons)],
                           columns= [args.orthogroup_column, 'Tree'])
        df_r.Tree.apply(lambda x: x if x=='*' else assign_id(x))
        df_r.Tree= df_r.Tree.apply(lambda x: x if x=='*' else get_nhx(x, name_attr= 'label', root= 1))

        opath= f'{args.output_prefix}.resolved_trees.tsv'
        df_r.to_csv(opath, sep= '\t', index= False)
        print(f'Resolved gene trees successfully writen to {opath}')

        # Write mu maps
        for idx, X in zip(gTrees.index, recons):
            if X!='*':
                text_mu= recon_2_text(X[1])
                opath= f'{args.recon_output_dir}{args.output_prefix}.{idx}.reconciliation_map.tsv'
                with open(opath, 'w') as F:
                    F.write(text_mu)
        print(f'Reconciliation maps successfully writen at {args.recon_output_dir}')

        # Write labeled species tree
        nhx_s= get_nhx(sTree, name_attr= 'species', root= 1)
        opath= f'{args.output_prefix}.labeled_species_tree.nhxx'
        with open(opath, 'w') as F:
            F.write(nhx_s)
        print(f'Indexed species tree successfully writen to {opath}')
#





















import pandas as pd
import os
import numpy as np
import networkx as nx
from itertools import chain
from tqdm import tqdm
tqdm.pandas()

from .common_tools import norm_path
from .in_out import create_cBMG, tl_digraph_from_pandas

####################
#                  #
# Some parameters  #
#                  #
####################

from .constants import _df_matches_cols, _default_f

####################
#                  #
# Parse functions  #
#                  #
####################

def parse_prt_best_hits(path, f_value):
    """
    Returns a DataFrame of best hits obtained from proteinortho files
    """
    print('Reading .proteinortho.tsv file and hits directory...')
    # Check if all files are ok
    prt_path= _check_files(path)
    # Load prt orthogroups
    df_prt= parse_prt_orthogroups(prt_path)
    # Load prt hits
    df_hits= parse_prt_hits(path, best_hit_dir= f'proteinortho_cache_{_get_prt_projectname(prt_path)}')
    # Identify best hit scores w.r.t. query_gene and target_species
    w_x_B= max_hit_scores(df_hits)
    # Apply dynamic threshold for best hit selection
    print('Selecting best hits by dynamic threshold...')
    df_best_hits= select_best_hits(df_hits, f_value, w_x_B)
    # Return the subset of hits for each orthogroup
    print('Filtering best hits by orthogroup...')
    F= lambda row: _get_orthogroup_hits(row, df_best_hits)
    return pd.concat( df_prt.progress_apply(F, axis= 1).values )

def parse_prt_orthogroups(orthogroups_file, clear_singletones= True):
    df_prt= pd.read_csv(orthogroups_file, sep= '\t')
    species_cols= list(map(lambda x: x.split('.')[0], df_prt.columns[3:]))
    df_prt.columns= ['Species', 'Genes', 'Alg.-Conn.']+species_cols
    df_prt[species_cols]= df_prt[species_cols].apply(lambda X: X.apply(lambda x: set(x.split(','))))
    df_prt.index.name= 'OG'
    if clear_singletones:
        df_prt= _clean_prt(df_prt)
    return df_prt

def parse_prt_hits(prt_files_path,
                   best_hit_dir= '.',
                   best_hit_file_ext= '.diamond',
                   Query_accession= 'Query_accession',
                   pd_params= dict(names= _df_matches_cols, sep= '\t'),
                  ):
    """
    Returns a DataFrame of hits obtained from proteinortho files
    """
    best_hit_dir= norm_path(norm_path(prt_files_path)+best_hit_dir)

    df= pd.concat((_read_bh_table(f'{best_hit_dir}/{file}',
                                 pd_params,
                                 Query_accession= Query_accession
                                 )
                   for file in os.listdir(best_hit_dir) if file.endswith(best_hit_file_ext)
                  ))
    return df

def _read_bh_table(file, pd_params, Query_accession):
    """
    Para que funcione, el archivo debe ser llamado algo así como: 
    H0.fa.vs.H10.fa.diamond
    """
    species= file.split('/')[-1].split('.')
    species= species[0], species[3]
    df= pd.read_csv(file, **pd_params)
    columns= list(df.columns)
    df['Query_species']= species[0]
    df['Target_species']= species[1]
    return df[ ['Query_species', 'Target_species']+columns ]

def max_hit_scores(df_hits,
                   Query_accession= 'Query_accession',
                   Target_species= 'Target_species',
                   Score= 'Bit_score',
                  ):
    return df_hits.groupby([Query_accession, Target_species])[Score].max()

def select_best_hits(df, f_value, w_x_B, a= 'Query_accession', b= 'Target_accession', species_b= 'Target_species'):
    """
    Obtain the subset of hits that pass the dynamic threshold of proteinortho.
    """
    is_best_hit= lambda row: row.Bit_score >= f_value * w_x_B[ row.Query_accession, row.Target_species ]
    mask= df.progress_apply(is_best_hit, axis= 1)
    return df.loc[mask]

def _get_orthogroup_hits(row, df_hits):
    """
    Input:
    - row: A row from an orthogroups table (.proteinortho.tsv)
    - df_hits: table of best hits (Query_gene -> Target_gene)
    """
    genes= set(chain.from_iterable(row))
    mask= (df_hits.Query_accession.isin(genes)) & (df_hits.Target_accession.isin(genes))
    og_hits= df_hits[ mask ]
    og_hits['OG']= row.name
    return og_hits

def _clean_prt(df_prt):
    mask= df_prt.Species > 1
    species= list(df_prt.columns[3:])
    return df_prt.loc[mask, species]

def read_symBets(prt_graph_path, prt_path):
    # Loads symBets (proteinortho-graph)
    symBets= pd.read_csv(prt_graph_path, sep='\t', comment= '#', names= ['a',
                                                                         'b',
                                                                         'evalue_ab',
                                                                         'bitscore_ab',
                                                                         'evalue_ba',
                                                                         'bitscore_ba',
                                                                        ])

    # Loads proteiortho table containng orthogroups
    df_prt= parse_prt_orthogroups(prt_path)

    # For each ortogroup obtain the orthology relations of all the genes in such orthogroup
    F= lambda x: _get_orthos(x, symBets)
    df_edges= pd.concat( list(df_prt.apply(F, axis= 1)) )

    return df_edges

def _get_orthos(row, symBets):
    orths= pd.concat((_search_orths(row, X, symBets) for X in row.index))
    orths['OG']= row.name
    return orths.set_index('OG').sort_index()

def _search_orths(row, species, symBets):
    mask= lambda x: symBets[ (symBets.a==x) | (symBets.b==x) ]
    return pd.concat(map(mask, row[species]))

####################
#                  #
# General functions#
#                  #
####################

def _check_files(path):
    prt_path= [x for x in os.listdir(path) if x.endswith('.proteinortho.tsv')]
    if len(prt_path)==1:
        prt_path= path + prt_path[0]
    else:
        raise ValueError(f'There is more than one ".proteinortho.tsv" file at {path}')
    return prt_path

def _get_prt_projectname(prt_file):
    return prt_file.split('/')[-1].split('.')[0]

####################
#                  #
# Graph functions  #
#                  #
####################

def parse_prt_project_bmgs(path, f_value):
    df= parse_prt_best_hits(path, f_value)
    X= tl_digraph_from_pandas(df, og_col= 'OG')
    return X

####################
#                  #
# Standalone usage #
#                  #
####################


def _parse(path, f_value, opath):
    df= parse_prt_best_hits(path, f_value)
    df.reset_index().to_csv(opath, sep='\t', index= False)
    print(f'Best hits successfully writen to {opath}')

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(prog= 'parse_prt',
                                     description='Converts proteinortho outputs to edges list of best hits.',
                                    )
    # Parameters for input graph
    ############################

    parser.add_argument('prt_path',
                        type= str,
                        help= 'Path to a directory containing proteinortho files. For more information see:'
                       )

    parser.add_argument('-o', '--output_prefix',
                        type= str,
                        required=False,
                        help= 'prefix used for output files (default "prt_revolutionhtl").',
                        default= 'prt_revolutionhtl',
                       )

    parser.add_argument('-f', '--f_value',
                        type= float,
                        required=False,
                        help= f'Number between 0 and 1 (default {_default_f}), defines the adaptative threshhold for best matches: f*max_bit_score. (see proteinortho paper for a deep explanation)',
                        default= _default_f,
                       )

    args= parser.parse_args()

    # Process data
    ##############
    args.prt_path= norm_path(args.prt_path)
    opath= args.output_prefix+'.best_hits.tsv'
    _parse(args.prt_path, args.f_value, opath)





from itertools import product, takewhile, combinations
import pandas as pd
import numpy as np
from collections.abc import Iterable
import networkx as nx
from Bio import Phylo
from io import StringIO

__forbidden_attributes_phylo= ['_color', 'clades']


#> Define some functions
########################

# Para árboles

def get_dad(tree, n):
    in_edgs= list(tree.in_edges(n))
    if len(in_edgs)==0:
        if n in tree:
            raise ValueError('Node "{}" has no dad'.format(n))
        else:
            raise ValueError('Node "{}" not in tree'.format(n))
    return in_edgs[0][0]

def get_path(tree, x0, x1):
    P= [x1]
    while( x1 != x0 ):
        x1= get_dad(tree, x1)
        P+= [x1]
    return P[::-1]

def are_the_same(X):
    x0= X[0]
    F= lambda x: x==x0
    return all(map(F, X))

def LCA(tree, nodes):
    if len(nodes)==0:
        raise ValueError(f'Parameter "nodes" must be a non-empty set.')
    P= [get_path(tree, tree.root, n) for n in nodes]
    L= min(map(len, P))
    P= [X[:L] for X in P]
    M= len(nodes)
    #X= list( takewhile(lambda x: are_the_same(P[:,x]) , range(L)) )[-1]    
    X= list( takewhile(lambda x: are_the_same( [P[i][x] for i in range(M)] ) , range(L)) )[-1]    
    return P[0][X]

def LCA_color(tree, colors, color_attr= 'color'):
    if len(colors)==0:
        raise ValueError(f'Parameter "colors" must be a non-empty set.')
    nodes= {x for x in tree if tree.nodes[x][color_attr] in colors}
    return LCA(tree, nodes)

def induced_leafs(tree, n):
    Q= [n]
    L= []
    while len(Q) > 0:
        n= Q.pop()
        sons= list(tree[n])
        if len(sons) == 0:
            L+= [n]
        else:
            Q+= sons
    return L

def get_inner_nodes(tree):
    return set(tree) - set(induced_leafs(tree, tree.root))

def is_leaf(tree, node):
    return len(tree[node]) == 0

def get_color2leafs(tree, color_attr= 'color'):
    leafs= induced_leafs(tree, tree.root)
    colors= set((tree.nodes[leaf][color_attr] for leaf in leafs))
    color_2_leaf= {color:set() for color in colors}
    for leaf in leafs:
        color_2_leaf[ tree.nodes[leaf][color_attr] ].add(leaf)
    return color_2_leaf

def get_symbolic_relations(T, sym, nAttr):
    edges= []
    for node in T:
        if T.nodes[node][nAttr]==sym:
            for x0,x1 in combinations(T[node], 2):
                edges+= list(product(induced_leafs(T, x0),
                                     induced_leafs(T, x1)
                                    ))
    return edges

def copy_tree(tree):
    root= tree.root
    tree= tree.copy()
    tree.root= root
    return tree

def subtree(T, root):
    nodes= BFS_graph(T, root)
    T1= nx.induced_subgraph(T, nodes)
    T1.root= root
    return T1

# Leafs overlap | Andrea
def are_overlapping(set_list):
    ret= False
    for s0,s1 in combinations(set_list, 2):
        if not s0.isdisjoint(s1):
            ret= True
            break
    return ret

def leafs_overlap(tree, node):
    if not is_leaf(tree, node):
        return are_overlapping([induced_colors(tree, child)
                                for child in tree[node]])
    else:
        raise ValueError(f'leafs overlap not defined for a leaf ({node})')


def assign_symbolic_date(tree):
    for node in tree:#BFS_graph(tree, tree.root):
        if not is_leaf(tree, node):
            if leafs_overlap(tree, node):
                tree.nodes[node]['color']= 'D'
            else:
                tree.nodes[node]['color']= 'S'

# Pasar a otros tipos de datos

def get_newick(T):
    newick= {}
    for node in nx.dfs_postorder_nodes(T):
        if len(T[node])==0 :
            newick[ node ]= 'spec' + node#nxTree.nodes[node]['label']
        else:
            newick[ node ]= '(' + ','.join([ newick[son] for son in T[node] ]) + ')'
    return newick[node]+';'

def __getNAme(node, idx, tree, label):
    attributes= {a:b for a,b in node.__dict__.items()
                 if b != None
                 and a not in __forbidden_attributes_phylo
                }
    n_label= attributes.get(label, None)
    if n_label==None:
        return idx, attributes
    if n_label in tree:
        raise ValueError('The label "{}" is not unique for every node'.format(n_label))
    return n_label, attributes

def make_directed(X):
    """
    Transforma un árbol con estrucutura de dato nx.Graph  a un objeto de networkx.Digraph sin ningun elemento de biopython
    """
    tree= nx.DiGraph()
    L= []
    for x0 in BFS_graph(X.nx_tree, X.bio_tree.root):
        L+= [x0]
        for x1 in X.nx_tree[x0]:
            if x1 not in L:
                w= X.nx_tree[x0][x1]['weight']
                tree.add_edge(x0, x1, weight= w)

    tree.root= X.bio_tree.root
    return tree


# Recorridos

def BFS_graph(tree, root):
    """
    Returns a list of nodes visited in a BFS travel

    tree: nx.Graph
    root: node in nx.Graph.nodes
    """
    Q= [root]
    L= []
    while len(Q) > 0:
        x= Q.pop(0)
        L+= [x]
        Q+= [neighbor for neighbor in tree[x]
             if neighbor not in L]
    return L

# auxiliares

def is_superColor(color):
    return isinstance(color, Iterable) and not type(color)==str

def correct_colors(tree):
    for n in tree:
        colorts= tree.nodes[n]['color']
        #if( type(colorts)==list ):
        #if isinstance(colorts, Iterable) :
        if is_superColor(colorts):
            tree.nodes[n]['color']= list(map(str, colorts))
        else:
            tree.nodes[n]['color']= str(colorts)
    tree.root= '0'
    return tree

def induced_colors(tree, node, color_attr= 'color'):
    sLeafs= set()
    for x in induced_leafs(tree, node):
        xcol= tree.nodes[x][color_attr]
        if is_superColor(xcol):
            sLeafs.update(xcol)
        else:
            sLeafs.add(xcol)
    return sLeafs






_df_matches_cols= ['Query_accession',
                  'Target_accession',
                        'Sequence_identity',
                        'Length',
                        'Mismatches',
                        'Gap_openings',
                        'Query_start',
                        'Query_end',
                        'Target_start',
                        'Target_end',
                        'E_value',
                        'Bit_score',
                       ]
_default_f= 0.95



from . import nxTree as nxt
import networkx as nx
from itertools import chain, product, combinations

def get_triplets(tree, event='event', color= 'color', root_event= 'S'):
    I= {} # Dictionary for induced leafs
    R= set() # set for tree triplets
    for x in nx.dfs_postorder_nodes(tree):
        if nxt.is_leaf(tree, x):
            I[x]= { tree.nodes[x][color] }
        else:
            if x!=tree.root:
                I[x]= set( chain.from_iterable((I[x1] for x1 in tree[x])) ) #### Acomodar esta linea: que no se calcule en la raíz
            if tree.nodes[x][event] == root_event:
                R.update( _get_triples_from_root(tree, x, I) )
    R= [_construct_triplet( X, color ) for X in R]
    return R

def _get_triplets_from_groups(tree, x_out, x_in, I):
    P= combinations( I[x_in], 2 )
    R= set(( (a, frozenset([b,c]))
             for a,(b,c) in product( I[x_out], P ) if a!=b and a!=c
          ))
    return R

def _get_triples_from_root(tree, node, I):
    R= set()
    for x0, x1 in combinations(tree[node], 2):
        R.update( _get_triplets_from_groups(tree, x0, x1, I) )
        R.update( _get_triplets_from_groups(tree, x1, x0, I) )
    return R

def _construct_triplet(topology, color):
    a, (b, c)= topology
    T= nx.DiGraph()
    T.root= 0
    T.add_nodes_from([0,1])
    for i,x in enumerate((a, b, c)):
        T.add_node(i+2, **{color: x})
    T.add_edges_from(( (0,1), (0,2), (1,3), (1,4)))
    return T



from .nxTree import copy_tree, is_leaf, induced_colors
from itertools import product, combinations
import networkx as nx
from bmgedit import BMGEditing
import pandas as pd
from tqdm import tqdm
tqdm.pandas()

def build_graph(G, method='louvain', objective='cost', color_attr= None):
    """
    ARG1: method - this param will be use to select the method
    """
    if color_attr != None:
        G1= G.copy()
        for x in G:
            G1.nodes[x]['color']= G.nodes[x][color_attr]
    else:
        G1= G

    solver = BMGEditing.BMGEditor(G1, binary=True)
    solver.build(method, objective)

    hug_free= solver.get_bmg(extract_triples_first=False)

    for x in G:
        hug_free.nodes[x]['accession']= G.nodes[x]['accession']
        hug_free.nodes[x]['species']= G.nodes[x]['species']

    return hug_free

def _addN(T,idx, tree, n):
    T.add_node(idx, **(tree.nodes[n].copy()))

def get_augmented_tree(tree, geneAttr, speciesAttr):
    """
    Constructs the color intersection graph
    Schaller et. al. 2020:
    Complete characterization of incorrect orthology assiments in BMGs
    """

    T= nx.DiGraph() #<
    T.root= 0
    _addN(T, 0, tree, 0) #<
    w_idx= 1 #<
    mu= {1 : 0} #< node x in tree --> corresponding dad in T

    for node in list(nx.dfs_preorder_nodes(tree, source= 1)):

        _addN(T, w_idx, tree, node) #<
        T.add_edge(mu[node], w_idx) #<
        for child in tree[node]:
            mu[child]= w_idx
        current= w_idx
        w_idx+= 1

        if (not is_leaf(tree, node)) and (node != tree.root):
            CIG= get_CIG(tree, node, speciesAttr)
            CC= list(nx.connected_components(CIG))
            if len(CC)>1:
                T.nodes[current][geneAttr]= 'S' #<
                for cc in (X for X in CC if len(X)>1):
                    _create_redundant_node(T, w_idx, current, tree, node, cc, mu, geneAttr, speciesAttr) #<
                    w_idx+= 1
            else:
                T.nodes[current][geneAttr]= 'P'
    T.nodes[T.root][geneAttr]= None
    return T

def get_CIG(tree, node, speciesAttr):
    """
    Color Intersection Graph
    """
    children= set(tree[node])
    CIG= nx.Graph()
    CIG.add_nodes_from(children)
    for v0,v1 in combinations(children, 2):
        colors0= induced_colors(tree, v0, speciesAttr)
        colors1= induced_colors(tree, v1, speciesAttr)
        if not colors0.isdisjoint(colors1):
            CIG.add_edge(v0, v1)
    return CIG

def _create_redundant_node(T, w_idx, current, tree, node, cc, mu, geneAttr, speciesAttr):
    T.add_node(w_idx, **{geneAttr: 'P', speciesAttr: None})
    T.add_edge(current, w_idx)
    for child in cc:
        mu[child]= w_idx

####################
#                  #
# Standalone usage #
#                  #
####################

if __name__ == "__main__":
    import argparse
    from .parse_prt import create_cBMG
    from .in_out import read_tl_digraph, write_digraphs_list

    parser = argparse.ArgumentParser(prog= 'hug_free',
                                     description='Converts a digraph into a BMG.',
                                    )

    # Parameters for input graph
    ############################

    parser.add_argument('edges_list',
                        type= str,
                        help= '.tsv file containing best matches. (See input formats)'
                       )

    parser.add_argument('-og', '--orthogroup',
                        type= str,
                        required=False,
                        help= 'Column in the edges_list specifying orthogroups.',
                        default= None,
                       )

    parser.add_argument('-f', '--filter_orthogroups',
                        type= str,
                        required=False,
                        help= '.tsv file describing which orthogroups are alrready cBMGs.',
                        default= None,
                       )
    

    parser.add_argument('-o', '--output_prefix',
                        type= str,
                        required=False,
                        help= 'prefix used for output files (default "prt_revolutionhtl").',
                        default= 'tl_project.hug_free',
                       )

    args= parser.parse_args()

    # Process data
    ##############

    print('Reading hit graphs...')
    G= read_tl_digraph(args.edges_list, og_col= 'OG')
    print('Computing hug-free graphs...')
    G1= G.progress_apply(lambda X: build_graph(X, color_attr= 'species'))
    write_digraphs_list(G, args.output_prefix, n_attrs= ['accession', 'species'])
    print(f'Hug-free graphs successfully writen to {args.output_prefix}.tsv')
